#ifndef VCACHE_INCLUDED
#define VCACHE_INCLUDED

#include "haxmath.h"

using namespace HAXMATH;

const int DEFAULT_VCACHE_SIZE = 20;

const int MAX_VCACHE_SIZE = 64;

//====================================================================
//====================================================================
template <class TIndexType>
DWORD FindVertexCount(DWORD faceCount, const TIndexType* pIndex)
{
 if (faceCount==0) return 0;
 
 DWORD nVertices = 0;

 const TIndexType* p=pIndex;
 for (DWORD i=0; i<faceCount*3; i++)
  {
   nVertices = max<DWORD>(nVertices, *p);
   p++;
  }

 return nVertices+1; 
}

//====================================================================
//====================================================================
//measure Post-TnL cache efficiency of cache with specified size
//Efficiency is number of cache misses per triangle - less is better.
template <class TIndexType>
float MeasurePostTnLCacheEfficiency(DWORD faceCount, const TIndexType* pIndex, DWORD VCacheSize = DEFAULT_VCACHE_SIZE)
{
 if (faceCount==0) return 0;

 assert(VCacheSize<MAX_VCACHE_SIZE);

 DWORD nVertices = FindVertexCount(faceCount, pIndex);
 
 bool* bInCache = new bool[nVertices];
 memset(bInCache,0,nVertices);
 
 DWORD cache[MAX_VCACHE_SIZE];
 
 for (DWORD i=0; i<VCacheSize; i++)
  {
   cache[i]=MAXDWORD;
  }
  
 DWORD cacheHead = 0;
 
 DWORD misses = 0;
 
 const TIndexType* p=pIndex;
 for (DWORD i=0; i<faceCount*3; i++)
  {
   if (bInCache[*p]==false) 
    {
     misses++;
     
     if (cache[cacheHead]!=MAXDWORD)
      {
       bInCache[cache[cacheHead]]=false;
      } 

     cache[cacheHead] = *p;
     bInCache[*p]=true;
     
     cacheHead++;
     if (cacheHead>VCacheSize-1) cacheHead=0;
    }
   
   p++;
  }
 
 delete[] bInCache; 
 
 return ((float)misses)/(faceCount);
}

//====================================================================
//====================================================================
//measure Pre-TnL cache efficiency with Post-TnL cache with specified size
//Efficiency is sum of distances between vertexes, missed in Post-TnL cache,
//divided by number of misses - less is better.
//Ideal is linear access, when new vertex is situated just after previous - 1.0f.
template <class TIndexType>
float MeasurePreTnLCacheEfficiency(DWORD faceCount, const TIndexType* pIndex, DWORD VCacheSize = DEFAULT_VCACHE_SIZE)
{
 if (faceCount==0) return 0;

 assert(VCacheSize<MAX_VCACHE_SIZE);

 DWORD nVertices = FindVertexCount(faceCount, pIndex);
 
 bool* bInCache = new bool[nVertices];
 memset(bInCache,0,nVertices);
 
 DWORD cache[MAX_VCACHE_SIZE];
 
 for (DWORD i=0; i<VCacheSize; i++)
  {
   cache[i]=MAXDWORD;
  }
  
 DWORD cacheHead = 0;
 
 float dist = 0;
 float misses = 0;
 DWORD lastIndex = pIndex[0];
 
 const TIndexType* p=pIndex;
 for (DWORD i=0; i<faceCount*3; i++)
  {
   if (bInCache[*p]==false) 
    {
     dist += fabs((float)*p - (float)lastIndex);
     misses += 1.0f;
     
     lastIndex = *p;
     
     if (cache[cacheHead]!=MAXDWORD)
      {
       bInCache[cache[cacheHead]]=false;
      } 

     cache[cacheHead] = *p;
     bInCache[*p]=true;
     
     cacheHead++;
     if (cacheHead>VCacheSize-1) cacheHead=0;
    }
   
   p++;
  }
 
 delete[] bInCache; 
 
 if (misses>0.0f)
  {
   float f = dist/misses;
   if (f<1.0f) f=1.0f; //fix float accuracy - 0.9999999
   return f;
  }
   else
  {
   return 0;
  } 
}

//====================================================================
//====================================================================
template <class TIndexType>
void OptimizeFacesD3DX(DWORD faceCount, const TIndexType* pIndex, TIndexType* newIB)
{
 assert(pIndex!=newIB);
 assert(sizeof(TIndexType)==2 || sizeof(TIndexType)==4);
 
 DWORD* faceRemap = new DWORD[faceCount];

 D3DXOptimizeFaces(pIndex,
                   faceCount,
                   FindVertexCount(faceCount, pIndex),
                   sizeof(TIndexType)==4,
                   faceRemap);
                   
 for (DWORD i=0; i<faceCount; i++)
  {
   newIB[i*3]=pIndex[faceRemap[i]*3];
   newIB[i*3+1]=pIndex[faceRemap[i]*3+1];
   newIB[i*3+2]=pIndex[faceRemap[i]*3+2];
  }     
  
 delete[] faceRemap;              
}


//====================================================================
//====================================================================
template <class TIndexType>
void CreateConnectivity(DWORD faceCount, DWORD nVertices, TIndexType* pIndex, DWORD* &connectivity, DWORD* &connectivityData)
{
 connectivity = new DWORD[nVertices+1];
 memset(connectivity, 0, (nVertices+1)*sizeof(DWORD));

 for (DWORD i=0; i<faceCount*3; i++)
  {
   assert(pIndex[i]<nVertices);
   assert(pIndex[i]>=0);
   connectivity[pIndex[i]]++;
  }
  
 DWORD j=0;
 
 for(DWORD i=0; i<nVertices+1; i++)
  {
   DWORD k = connectivity[i];
   connectivity[i]+=j;
   j+=k;
  } 
 
 assert(j==faceCount*3);
 
 connectivityData = new DWORD[faceCount*3];
 
 connectivity[nVertices] = connectivity[nVertices-1];
 
 for (DWORD i=0; i<faceCount*3; i++)
  {
   TIndexType index = pIndex[i];
   connectivity[index]--;
   connectivityData[connectivity[index]]= i/3;
  }
}


//====================================================================
//====================================================================
//Optimize for FIFO cache of specified size.
//It is better to underestimate then overestimate real cache size,
//because produced order will trash caches of lower size.
//see http://developer.nvidia.com/object/vertex_cache_opt.html
//Surprisingly this algorithm produces worst results.
//The problem seems to be that algorithm is trying to keep 
//only last two vertexes in cache, not accounting for vertex position in cache.
//LRU algorithm is also trying to build long strips, instead of selecting
//best-score face at each step.
//newIB should be allocated
template <class TIndexType>
void OptimizeFacesFIFO(DWORD faceCount, const TIndexType* pIndex, TIndexType* newIB, DWORD FIFO_CACHE_SIZE = DEFAULT_VCACHE_SIZE)
{
 if (faceCount==0) return;

 assert(FIFO_CACHE_SIZE<=MAX_VCACHE_SIZE);

 DWORD nVertices = FindVertexCount(faceCount, pIndex);
 
 bool* bInCache = new bool[nVertices];
 memset(bInCache,0,nVertices);
 
 DWORD misses = 0;

 DWORD cache[MAX_VCACHE_SIZE];
 
 for (DWORD i=0; i<FIFO_CACHE_SIZE; i++)
  {
   cache[i]=MAXDWORD;
  }
  
 DWORD cacheHead = 0;
 
 DWORD* connectivity;
 DWORD* connectivityData;
 CreateConnectivity(faceCount, nVertices, pIndex, connectivity, connectivityData);
 
 //holds new position of face [index]
 DWORD* faceRemap = new DWORD[faceCount];
 for (DWORD i=0; i<faceCount; i++)
  {
   faceRemap[i]=MAXDWORD;
  }
                       
 DWORD count = 0;                       
 
 while (count<faceCount)
  {
   DWORD best = MAXDWORD;
   DWORD bestCost;
   
   for (DWORD i=0; i<faceCount; i++)
    {
     if (faceRemap[i]!=MAXDWORD) continue;
     
     DWORD thisCost = 0;
     
     const TIndexType* tri = &pIndex[i*3];
     
     bool pushFlags[3];
     for (DWORD l=0; l<3; l++)
      {
       pushFlags[l] = bInCache[tri[l]];
      }
     
     for (DWORD l=0; l<3; l++)
      {
       
       if (bInCache[tri[l]]==false) 
        {
         thisCost++;
         //we are pushing a vertex out from cache, mark it 'not in cache' temporarily
         DWORD k = (cacheHead+thisCost) % FIFO_CACHE_SIZE;
         DWORD pushOutIndex = cache[k];
         
         if (pushOutIndex!=MAXDWORD)
          {
           bInCache[pushOutIndex]=false;
          } 

         //we are pushing a vertex into cache
         bInCache[tri[l]]=true;
        } 
      }  
        
     //by adding this face, we are pushing out pushOutCount vertexes from the cache
     DWORD pushOutCount = thisCost;

     //restore flags
     for (DWORD l=0; l<pushOutCount; l++)
      {
       DWORD k = (cacheHead+l+1) % FIFO_CACHE_SIZE;

       DWORD pushOutIndex = cache[k];
       if (pushOutIndex==MAXDWORD) continue;

       bInCache[pushOutIndex]=true;
      }    

     for (DWORD l=0; l<3; l++)
      {
       bInCache[tri[l]] = pushFlags[l];
      } 
     
     //for every vertex, pushed out of cache,
     //for every unprocessed triangle, using this vertex,
     //add cost for every vertex of this triangle which is in cache
          
     for (DWORD j=0; j<pushOutCount; j++)
      {
       DWORD k = (cacheHead+j+1) % FIFO_CACHE_SIZE;
       DWORD pushOutIndex = cache[k];
       if (pushOutIndex==MAXDWORD) continue;
       
       for (DWORD l=connectivity[pushOutIndex]; l<connectivity[pushOutIndex+1]; l++)
        {
         DWORD t = connectivityData[l];

         const TIndexType* tv = &pIndex[t*3];

         assert(tv[0]==pushOutIndex || tv[1]==pushOutIndex || tv[2]==pushOutIndex);

         if (t==i) continue;
         if (faceRemap[t]!=MAXDWORD) continue;
         
         if (bInCache[tv[0]]) thisCost++;
         if (bInCache[tv[1]]) thisCost++;
         if (bInCache[tv[2]]) thisCost++;
        }
      }
            
      
     if (best==MAXDWORD || thisCost<bestCost)
      {
       best=i;
       bestCost=thisCost;
       if (bestCost==0) break;
      } 
    }
    
   //store face [best]
   faceRemap[best]=count;
   
   //push vertexes into cache
   
   for (DWORD j=0; j<3; j++)
    {
     const TIndexType* tri = &pIndex[best*3];
     
     if (bInCache[tri[j]]==false)
      {
       DWORD k = (cacheHead+1) % FIFO_CACHE_SIZE;
       DWORD pushOutIndex = cache[k];
       if (pushOutIndex!=MAXDWORD) 
        {
         assert(bInCache[pushOutIndex]==true);
         bInCache[pushOutIndex]=false;
        } 
       
       cache[k]=tri[j];
       assert(bInCache[tri[j]]==false);
       bInCache[tri[j]]=true;
       cacheHead=k;       
       
       misses++;
      } 
    } 
    
   count++;
  }
  
  for (DWORD i=0; i<faceCount; i++)
   {
    newIB[faceRemap[i]*3]=pIndex[i*3];
    newIB[faceRemap[i]*3+1]=pIndex[i*3+1];
    newIB[faceRemap[i]*3+2]=pIndex[i*3+2];
   }     
 
 delete[] faceRemap;
 delete[] connectivity;
 delete[] connectivityData;
 delete[] bInCache; 
 
 float f =((float)misses)/faceCount;
 float f1=f;
}




extern const float __declspec(selectany) FindVertexScore_CacheDecayPower = 1.5f;
extern const float __declspec(selectany) FindVertexScore_LastTriScore = 0.75f;
extern const float __declspec(selectany) FindVertexScore_ValenceBoostScale = 2.0f;
extern const float __declspec(selectany) FindVertexScore_ValenceBoostPower = 0.5f;

//====================================================================
//====================================================================
float __inline FindVertexScore(DWORD cachePos, DWORD activeFaceCount, DWORD MaxSizeVertexCache)
{
 if (activeFaceCount == 0)
    {
     // No tri needs this vertex!
     return -1.0f;
    }

 float Score = 0.0f;

 DWORD CachePosition = cachePos;

 if ( CachePosition == MAXDWORD )
   {
     // Vertex is not in FIFO cache - no score.
   }
    else
   {
    if ( CachePosition < 3 )
     {
      // This vertex was used in the last triangle,
      // so it has a fixed score, whichever of the three
      // it's in. Otherwise, you can get very different
      // answers depending on whether you add
      // the triangle 1,2,3 or 3,1,2 - which is silly.

      Score = FindVertexScore_LastTriScore;
     }
      else
     {
      assert ( CachePosition < MaxSizeVertexCache );

      // Points for being high in the cache.

      const float Scaler = 1.0f / ( MaxSizeVertexCache - 3 );

      Score = 1.0f - ( CachePosition - 3 ) * Scaler;

      Score = powf ( Score, FindVertexScore_CacheDecayPower );
     }
   }

 // Bonus points for having a low number of tris still to 
 // use the vert, so we get rid of lone verts quickly.

 float ValenceBoost = powf ((float)activeFaceCount,  -FindVertexScore_ValenceBoostPower );

 Score += FindVertexScore_ValenceBoostScale * ValenceBoost;
 return Score;
}

 
 
//====================================================================
//====================================================================
//Optimize for LRU cache of specified size.
//see "Linear-Speed Vertex Cache Optimization" by Tom Forsyth
//http://home.comcast.net/~tom_forsyth/papers/fast_vert_cache_opt.html.
//This algorithm produces best results.
//Not optimized - suits my needs as is.
//newIB should be allocated
template <class TIndexType>
void OptimizeFacesLRU(DWORD faceCount, const TIndexType* pIndex, TIndexType* newIB, DWORD LRU_CACHE_SIZE = MAX_VCACHE_SIZE)
{
 if (faceCount==0) return;

 assert(LRU_CACHE_SIZE<=MAX_VCACHE_SIZE);

 DWORD nVertices = FindVertexCount(faceCount, pIndex);
 
 DWORD* cachePos = new DWORD[nVertices];
 memset(cachePos,MAXBYTE,nVertices*sizeof(DWORD));
 
 DWORD cache[MAX_VCACHE_SIZE+3];
 
 for (DWORD i=0; i<LRU_CACHE_SIZE; i++)
  {
   cache[i]=MAXDWORD;
  }
  
 //holds new position of face [index]
 DWORD* faceRemap = new DWORD[faceCount];
 for (DWORD i=0; i<faceCount; i++)
  {
   faceRemap[i]=MAXDWORD;
  }
 
 //number of unprocessed faces using vertex 
 DWORD* activeFaceCount = new DWORD[nVertices];
 
 memset(activeFaceCount, 0, nVertices*sizeof(DWORD));
 
 for (DWORD i = 0; i<faceCount*3; i++)
  {
   activeFaceCount[pIndex[i]]++;
  }
  
 float* vertexScore = new float[nVertices];
 
 for (DWORD i=0; i<nVertices; i++)
  {
   vertexScore[i] = FindVertexScore(cachePos[i], activeFaceCount[i], LRU_CACHE_SIZE);
  } 
                       
 DWORD count = 0;                       
 
 while (count<faceCount)
  {
   DWORD best = MAXDWORD;
   float bestScore;
   
   for (DWORD i=0; i<faceCount; i++)
    {
     if (faceRemap[i]!=MAXDWORD) continue;
     
     const TIndexType* tri = &pIndex[i*3];
     
     float thisScore = vertexScore[tri[0]] + vertexScore[tri[1]] + vertexScore[tri[2]];
      
     if (best==MAXDWORD || thisScore>bestScore)
      {
       best=i;
       bestScore=thisScore;
      } 
    }
   
   //store face [best]
   faceRemap[best]=count;

   const TIndexType* tri = &pIndex[best*3];
   
   assert(activeFaceCount[tri[0]]>0);
   assert(activeFaceCount[tri[1]]>0);
   assert(activeFaceCount[tri[2]]>0);
   
   activeFaceCount[tri[0]]--;
   activeFaceCount[tri[1]]--;
   activeFaceCount[tri[2]]--;
   
   DWORD cp[3];
   
   cp[0]=cachePos[tri[0]];
   cp[1]=cachePos[tri[1]];
   cp[2]=cachePos[tri[2]];
   
   //remove cachepos tags
   for (DWORD j=0; j<LRU_CACHE_SIZE; j++)
    {
     if (cache[j]==MAXDWORD) continue;
     cachePos[cache[j]]=MAXDWORD;
    }
    
   DWORD pushOutCount = 0;
   
   //remove faces of this tri from cache 
   for (DWORD j=0;  j<3; j++)
    {
     if (cp[j]!=MAXDWORD)
      {
       memmove(&cache[cp[j]], &cache[cp[j]+1], (LRU_CACHE_SIZE+3-(cp[j]+1))*sizeof(DWORD));
       
       for (DWORD k=j+1; k<3; k++)
        {
         if (cp[k]!=MAXDWORD && cp[k]>=cp[j]+1) cp[k]--;
        }
      }
       else
      {
       pushOutCount++; 
      }
    } 
   
   //insert vertices of best triangle into head of cache 
   memmove(&cache[3],&cache[0], LRU_CACHE_SIZE*sizeof(DWORD));
   
   cache[0]=tri[2];
   cache[1]=tri[1];
   cache[2]=tri[0];
   //set cachepos tags
   
   for (DWORD j=0; j<LRU_CACHE_SIZE; j++)
    {
     if (cache[j]==MAXDWORD) continue;
     cachePos[cache[j]]=j;
    }
   
   //update scores of vertices in cache and vertices, pushed out
   
   for (DWORD j=0; j<LRU_CACHE_SIZE+pushOutCount; j++)
    {
     DWORD vIndex = cache[j];
     if (vIndex==MAXDWORD) continue;
     vertexScore[vIndex]=FindVertexScore(cachePos[vIndex], activeFaceCount[vIndex], LRU_CACHE_SIZE);
    }
   
   count++;
  }
  
  for (DWORD i=0; i<faceCount; i++)
   {
    newIB[faceRemap[i]*3]=pIndex[i*3];
    newIB[faceRemap[i]*3+1]=pIndex[i*3+1];
    newIB[faceRemap[i]*3+2]=pIndex[i*3+2];
   }     
 
 delete[] faceRemap;
 delete[] cachePos; 
 delete[] activeFaceCount;
 delete[] vertexScore;
}


//====================================================================
//====================================================================
//Rearrange vertices for linear access
//vertexRemap[n] - new place of old vertex[n]
//vertexRemap should be allocated to DWORD[numberOfVertices]
//vertexRemap[n] will be MAXDWORD if vertex n is isolated
template <class TIndexType>
void OptimizeVertices(DWORD faceCount, const TIndexType* pIndex, DWORD* vertexRemap)
{
 DWORD nVertices = FindVertexCount(faceCount, pIndex);
 
 memset(vertexRemap,MAXBYTE,nVertices*sizeof(DWORD));
 
 DWORD count = 0;
 
 for (DWORD i=0; i<faceCount*3; i++)
  {
   TIndexType vIndex = pIndex[i];
   
   if (vertexRemap[vIndex]==MAXDWORD)
    {
     vertexRemap[vIndex]=count;
     count++;
    }
  }
}


#endif VCACHE_INCLUDED